from .dtw import *
from .dtwapprox import fastdtw, mrmsdtw
from .audiotools import *
from .alignmenttools import get_csm, get_path_cost, get_alignment_row_col_dists

from ._version import __version__



"""
Provides an interface to CUDA for running a parallel version
of the diagonal DTW algorithm
"""

import numpy as np
from .alignmenttools import get_diag_len, get_diag_indices, update_alignment_metadata
import warnings

DTW_Step_ = None
DTW_GPU_Initialized = False
DTW_GPU_Failed = False

def init_gpu():
    s = """
    __global__ void DTW_Diag_Step(float* d0, float* d1, float* d2, float* csm0, float* csm1, float* csm2, float* X, float* Y, int dim, int diagLen, int* box, int reverse, int i, int debug, float* U, float* L, float* UL, float* S) {
        //Other local variables
        int i1, i2, j1, j2; // Endpoints of the diagonal
        int thisi, thisj; // Current indices on the diagonal
        // Optimal score and particular score for up/right/left
        float score, left, up, diag; 
        int idx = threadIdx.x + blockIdx.x*blockDim.x;
        int xi, yj;

        //Process each diagonal
        score = -1;
        if (idx < diagLen) {
            // Figure out indices in X and Y on diagonal
            int M = box[1] - box[0] + 1;
            int N = box[3] - box[2] + 1;
            i1 = i;
            j1 = 0;
            if (i >= M) {
                i1 = M-1;
                j1 = i - (M-1);
            }
            j2 = i;
            i2 = 0;
            if (j2 >= N) {
                j2 = N-1;
                i2 = i - (N-1);
            }
            thisi = i1 - idx;
            thisj = j1 + idx;
            
            if (thisi >= i2 && thisj <= j2) {
                xi = thisi;
                yj = thisj;
                if (reverse == 1) {
                    xi = M-1-xi;
                    yj = N-1-yj;
                }
                xi += box[0];
                yj += box[2];
                // Step 1: Update csm2
                csm2[idx] = 0.0;
                for (int d = 0; d < dim; d++) {
                    float diff = X[xi*dim+d] - Y[yj*dim+d];
                    csm2[idx] += diff*diff;
                }
                csm2[idx] = sqrt(csm2[idx]);
                

                // Step 2: Figure out the optimal cost
                if (thisi == 0 && thisj == 0) {
                    score = 0;
                    if (debug == -1) {
                        S[0] = 0;
                        U[0] = -1;
                        L[0] = -1;
                        UL[0] = -1;
                    }
                }
                else {
                    left = -1;
                    up = -1;
                    diag = -1;
                    if (j1 == 0) {
                        if (idx > 0) {
                            left = d1[idx-1] + csm1[idx-1];
                        }
                        if (idx > 0 && thisi > 0) {
                            diag = d0[idx-1] + csm0[idx-1];
                        }
                        if (thisi > 0) {
                            up = d1[idx] + csm1[idx];
                        }
                    }
                    else if (i1 == M-1 && j1 == 1) {
                        left = d1[idx] + csm1[idx];
                        if (thisi > 0) {
                            diag = d0[idx] + csm0[idx];
                            up = d1[idx+1] + csm1[idx+1];
                        }
                    }
                    else if (i1 == M-1 && j1 > 1) {
                        left = d1[idx] + csm1[idx];
                        if (thisi > 0) {
                            diag = d0[idx+1] + csm0[idx+1];
                            up = d1[idx+1] + csm1[idx+1];
                        }
                    }
                    if (left > -1) {
                        score = left;
                    }
                    if (up > -1 && (up < score || score == -1)) {
                        score = up;
                    }
                    if (diag > -1 && (diag < score || score == -1)) {
                        score = diag;
                    }
                    if (debug == 1) {
                        U[thisi*N + thisj] = up;
                        L[thisi*N + thisj] = left;
                        UL[thisi*N + thisj] = diag;
                        S[thisi*N + thisj] = score;
                    }
                }
            }
            d2[idx] = score;
        }
    }
    """
    global DTW_GPU_Initialized
    if not DTW_GPU_Initialized:
        try:
            import pycuda.autoinit
            from pycuda.compiler import SourceModule
            mod = SourceModule(s)
            global DTW_Step_
            DTW_Step_ = mod.get_function("DTW_Diag_Step")
            DTW_GPU_Initialized = True
        except Exception as e:
            global DTW_GPU_Failed
            DTW_GPU_Failed = True
            warnings.warn("Unable to compile GPU kernel {}".format(e))

def dtw_diag_gpu(X, Y, k_save = -1, k_stop = -1, box = None, reverse=False, debug=False, metadata=None):
    """
    Compute dynamic time warping between two time-ordered
    point clouds in Euclidean space, using CUDA on the back end
    Parameters
    ----------
    X: ndarray(M, d)
        An M-dimensional Euclidean point cloud
    Y: ndarray(N, d)
        An N-dimensional Euclidean point cloud
    k_save: int
        Index of the diagonal d2 at which to save d0, d1, and d2
    k_stop: int
        Index of the diagonal d2 at which to stop computation
    debug: boolean
        Whether to save the accumulated cost matrix
    metadata: dictionary
        A dictionary for storing information about the computation
    """
    assert(X.shape[1] == Y.shape[1])
    global DTW_GPU_Initialized
    if not DTW_GPU_Initialized:
        init_gpu()
    if DTW_GPU_Failed:
        return
    import pycuda.gpuarray as gpuarray
    if not metadata:
        metadata = {}
    if not 'XGPU' in metadata:
        metadata['XGPU'] = gpuarray.to_gpu(np.array(X, dtype=np.float32))
    XGPU = metadata['XGPU']
    if not 'YGPU' in metadata:
        metadata['YGPU'] = gpuarray.to_gpu(np.array(Y, dtype=np.float32))
    YGPU = metadata['YGPU']
    if not box:
        box = [0, X.shape[0]-1, 0, Y.shape[0]-1]
    M = box[1] - box[0] + 1
    N = box[3] - box[2] + 1
    box_gpu = gpuarray.to_gpu(np.array(box, dtype=np.int32))
    reverse = np.array(reverse, dtype=np.int32)

    diagLen = np.array(min(M, N), dtype = np.int32)
    threadsPerBlock = min(diagLen, 512)
    gridSize = int(np.ceil(diagLen/float(threadsPerBlock)))
    threadsPerBlock = np.array(threadsPerBlock, dtype=np.int32)

    d0 = gpuarray.to_gpu(np.zeros(diagLen, dtype=np.float32))
    d1 = gpuarray.to_gpu(np.zeros(diagLen, dtype=np.float32))
    d2 = gpuarray.to_gpu(np.zeros(diagLen, dtype=np.float32))
    csm0 = gpuarray.zeros_like(d0)
    csm1 = gpuarray.zeros_like(d0)
    csm2 = gpuarray.zeros_like(d0)
    csm0len = diagLen
    csm1len = diagLen
    csm2len = diagLen

    if debug:
        U = gpuarray.to_gpu(np.zeros((M, N), dtype=np.float32))
        L = gpuarray.to_gpu(np.zeros((M, N), dtype=np.float32))
        UL = gpuarray.to_gpu(np.zeros((M, N), dtype=np.float32))
        S = gpuarray.to_gpu(np.zeros((M, N), dtype=np.float32))
    else:
        U = gpuarray.to_gpu(np.zeros(1, dtype=np.float32))
        L = gpuarray.to_gpu(np.zeros(1, dtype=np.float32))
        UL = gpuarray.to_gpu(np.zeros(1, dtype=np.float32))
        S = gpuarray.to_gpu(np.zeros(1, dtype=np.float32))

    res = {}
    for k in range(M+N-1):
        DTW_Step_(d0, d1, d2, csm0, csm1, csm2, XGPU, YGPU, np.array(X.shape[1], dtype=np.int32), diagLen, box_gpu, np.array(reverse, dtype=np.int32), np.array(k, dtype=np.int32), np.array(int(debug), dtype=np.int32), U, L, UL, S, block=(int(threadsPerBlock), 1, 1), grid=(gridSize, 1))
        csm2len = get_diag_len(box, k)
        update_alignment_metadata(metadata, csm2len)
        if k == k_save:
            res['d0'] = d0.get()
            res['csm0'] = csm0.get()[0:csm0len]
            res['d1'] = d1.get()
            res['csm1'] = csm1.get()[0:csm1len]
            res['d2'] = d2.get()
            res['csm2'] = csm2.get()[0:csm2len]
        if k == k_stop:
            break
        if k < M+N-2:
            # Rotate buffers
            temp = d0
            d0 = d1
            d1 = d2
            d2 = temp
            temp = csm0
            csm0 = csm1
            csm1 = csm2
            csm2 = temp
            temp = csm0len
            csm0len = csm1len
            csm1len = csm2len
            csm2len = temp
    res['cost'] = d2.get()[0] + csm2.get()[0]
    if debug:
        res['U'] = U.get()
        res['L'] = L.get()
        res['UL'] = UL.get()
        res['S'] = S.get()
    return res


import pytest
import numpy as np
import linmdtw

def get_pcs(N):
    t1 = np.linspace(0, 1, N)
    t2 = t1**2
    X = np.zeros((N, 2), dtype=np.float32)
    X[:, 0] = np.cos(2*np.pi*t1)
    X[:, 1] = np.sin(4*np.pi*t1)
    Y = np.zeros_like(X)
    Y[:, 0] = np.cos(2*np.pi*t2)
    Y[:, 1] = np.sin(4*np.pi*t2)
    return X, Y

class TestDTWApprox:

    def test_fasdtw(self):
        """
        Show that the fastdtw error is inversely correlated with the search
        radius
        """
        X, Y = get_pcs(1000)
        path1 = np.array(linmdtw.dtw_brute_backtrace(X, Y))
        path2 = np.array(linmdtw.fastdtw(X, Y, 50))
        path3 = np.array(linmdtw.fastdtw(X, Y, 5))
        err1 = linmdtw.get_alignment_row_col_dists(path1, path2)
        err2 = linmdtw.get_alignment_row_col_dists(path1, path3)
        assert(np.mean(err1) <= np.mean(err2))
        

    def test_mrmsdtw(self):
        """
        Test that the error is monotonically increasing when the memory
        is decreased in mrmsdtw
        """
        X, Y = get_pcs(1000)
        path1 = linmdtw.dtw_brute_backtrace(X, Y)
        path1 = np.array(path1)
        path2 = linmdtw.mrmsdtw(X, Y, tau=10**4)
        path2 = np.array(path2)
        path3 = linmdtw.mrmsdtw(X, Y, tau=10**2)
        path3 = np.array(path3)
        err1 = linmdtw.get_alignment_row_col_dists(path1, path2)
        err2 = linmdtw.get_alignment_row_col_dists(path1, path3)
        assert(np.mean(err1) <= np.mean(err2))
        assert(linmdtw.get_path_cost(X, Y, path2) < linmdtw.get_path_cost(X, Y, path3))

    def test_mrmsdtw_refine(self):
        """
        Test that error is lower after refinement
        """
        X, Y = get_pcs(1000)
        path1 = linmdtw.dtw_brute_backtrace(X, Y)
        path1 = np.array(path1)
        path2 = linmdtw.mrmsdtw(X, Y, tau=10**3, refine=True)
        path2 = np.array(path2)
        path3 = linmdtw.mrmsdtw(X, Y, tau=10**3, refine=False)
        path3 = np.array(path3)
        err1 = linmdtw.get_alignment_row_col_dists(path1, path2)
        err2 = linmdtw.get_alignment_row_col_dists(path1, path3)
        assert(np.mean(err1) <= np.mean(err2))
        assert(linmdtw.get_path_cost(X, Y, path2) < linmdtw.get_path_cost(X, Y, path3))


import numpy as np
from numba import jit

def get_diag_len(box, k):
    """
    Return the number of elements in this particular diagonal
    Parameters
    ----------
    MTotal: int
        Total number of samples in X
    NTotal: int
        Total number of samples in Y
    k: int
        Index of the diagonal
    Returns
    -------
    Number of elements
    """
    M = box[1] - box[0] + 1
    N = box[3] - box[2] + 1
    starti = k
    startj = 0
    if k >= M:
        starti = M-1
        startj = k - (M-1)
    endj = k
    endi = 0
    if k >= N:
        endj = N-1
        endi = k - (N-1)
    return endj-startj+1


def get_diag_indices(MTotal, NTotal, k, box = None, reverse=False):
    """
    Compute the indices on a diagonal into indices on an accumulated
    distance matrix
    Parameters
    ----------
    MTotal: int
        Total number of samples in X
    NTotal: int
        Total number of samples in Y
    k: int
        Index of the diagonal
    box: list [XStart, XEnd, YStart, YEnd]
        The coordinates of the box in which to search
    Returns
    -------
    i: ndarray(dim)
        Row indices
    j: ndarray(dim)
        Column indices
    """
    if not box:
        box = [0, MTotal-1, 0, NTotal-1]
    M = box[1] - box[0] + 1
    N = box[3] - box[2] + 1
    starti = k
    startj = 0
    if k > M-1:
        starti = M-1
        startj = k - (M-1)
    i = np.arange(starti, -1, -1)
    j = startj + np.arange(i.size)
    dim = np.sum(j < N) # Length of this diagonal
    i = i[0:dim]
    j = j[0:dim]
    if reverse:
        i = M-1-i
        j = N-1-j
    i += box[0]
    j += box[2]
    return i, j

def update_alignment_metadata(metadata = None, newcells = 0):
    """
    Add new amount of cells to the total cells processed,
    and print out a percentage point if there's been progress
    Parameters
    ----------
    newcells: int
        The number of new cells that are being processed
    metadata: dictionary
        Dictionary with 'M', 'N', 'totalCells', all ints
        and 'timeStart'
    """
    if metadata:
        if 'M' in metadata and 'N' in metadata and 'totalCells' in metadata:
            import time
            denom = metadata['M']*metadata['N']
            before = np.floor(50*metadata['totalCells']/denom)
            metadata['totalCells'] += newcells
            after = np.floor(50*metadata['totalCells']/denom)
            if after > before:
                print("Parallel Alignment {}% ".format(before), end='')
                if 'timeStart' in metadata:
                    print("Elapsed time: {}".format(time.time()-metadata['timeStart']))

def get_csm(X, Y):
    """
    Return the Euclidean cross-similarity matrix between X and Y
    Paramters
    ---------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    Returns
    -------
    D: ndarray(M, N)
        The cross-similarity matrix
    """
    XSqr = np.sum(X**2, 1)
    YSqr = np.sum(Y**2, 1)
    C = XSqr[:, None] + YSqr[None, :] - 2*X.dot(Y.T)
    C[C < 0] = 0
    return np.sqrt(C)

def get_ssm(X):
    """
    Return the SSM between all rows of a time-ordered Euclidean point cloud X
    Paramters
    ---------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Returns
    -------
    D: ndarray(M, M)
        The self-similarity matrix
    """
    return get_csm(X, X)

def get_path_cost(X, Y, path):
    """
    Return the cost of a warping path that matches two Euclidean 
    point clouds
    Paramters
    ---------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    P1: ndarray(K, 2)
        Warping path
    """
    x = X[path[:, 0], :]
    y = Y[path[:, 1], :]
    ds = np.sqrt(np.sum((x-y)**2, 1))
    return np.sum(ds)

def make_path_strictly_increase(path):
    """
    Given a warping path, remove all rows that do not
    strictly increase from the row before
    """
    toKeep = np.ones(path.shape[0])
    i0 = 0
    for i in range(1, path.shape[0]):
        if np.abs(path[i0, 0] - path[i, 0]) >= 1 and np.abs(path[i0, 1] - path[i, 1]) >= 1:
            i0 = i
        else:
            toKeep[i] = 0
    return path[toKeep == 1, :]

def get_alignment_area_dist(P1, P2, doPlot = False):
    """
    Compute area-based alignment error.  Assume that the 
    warping paths are on the same grid
    Parameters
    ----------
    ndarray(M, 2) pP1
        First warping path
    ndarray(N, 2) pP2
        Second warping path
    doPlot: boolean
        Whether to draw a plot showing the enclosed area
    Returns
    -------
    score: float
        The area score
    """
    import scipy.sparse as sparse
    M = np.max(P1[:, 0])
    N = np.max(P1[:, 1])
    A1 = sparse.lil_matrix((M, N))
    A2 = sparse.lil_matrix((M, N))
    for i in range(P1.shape[0]):
        [ii, jj] = [P1[i, 0], P1[i, 1]]
        [ii, jj] = [min(ii, M-1), min(jj, N-1)]
        A1[ii, jj::] = 1.0
    for i in range(P2.shape[0]):
        [ii, jj] = [P2[i, 0], P2[i, 1]]
        [ii, jj] = [min(ii, M-1), min(jj, N-1)]
        A2[ii, jj::] = 1.0
    A = np.abs(A1 - A2)
    dist = np.sum(A)/(M + N)
    if doPlot:
        import matplotlib.pyplot as plt
        plt.imshow(A.toarray())
        plt.scatter(P1[:, 1], P1[:, 0], 5, 'c', edgecolor = 'none')
        plt.scatter(P2[:, 1], P2[:, 0], 5, 'r', edgecolor = 'none')
        plt.title("Dist = %g"%dist)
    return dist

def get_alignment_cell_dists(P1, P2):
    """
    Return the L1 distances between each point on the warping path
    P2 to the closest point on the warping path P1
    Parameters
    ----------
    P1: ndarray(M, 2)
        Ground truth warping path
    P2: ndarray(N, 2)
        Test warping path
    Returns
    -------
    dists: ndarray(N)
        L1 distances of each point in the second warping path
        to their closest points in p1,
    """
    from sklearn.neighbors import KDTree, DistanceMetric
    tree = KDTree(P1, leaf_size=10)
    dists, _ = tree.query(P2, k=1)
    return dists.flatten()


@jit(nopython=True)
def get_alignment_row_dists(P1, P2):
    """
    For each point in the first path, record the distance
    of the closest point in the same row on the second path
    Parameters
    ----------
    P1: ndarray(M, 2)
        Ground truth warping path
    P2: ndarray(N, 2)
        Test warping path
    Returns
    -------
    dists: ndarray(M)
        The distances
    """
    i = 0
    k = 0
    dists = np.zeros(P1.shape[0])
    i2 = 0
    for i1 in range(P1.shape[0]):
        # Move along P2 until it's at the same row
        while P2[i2, 0] != P1[i1, 0]:
            i2 += 1
        # Check all entries of P2 that have the same row
        mindist = abs(P2[i2, 1] - P1[i1, 1])
        k = i2+1
        while k < P2.shape[0] and P2[k, 0] == P1[i1, 0]:
            mindist = min(mindist, abs(P2[k, 1]-P1[i1, 1]))
            k += 1
        dists[i1] = mindist
    return dists

def get_alignment_row_col_dists(P1, P2):
    """
    For each point in the first path, record the distance
    of the closest point in the same row on the second path,
    and vice versa.  Then repeat this along the columns
    Parameters
    ----------
    P1: ndarray(M, 2)
        Ground truth warping path
    P2: ndarray(N, 2)
        Test warping path
    Returns
    -------
    dists: ndarray(2M+2N)
        The distances
    """
    dists11 = get_alignment_row_dists(P1, P2)
    dists12 = get_alignment_row_dists(P2, P1)
    dists21 = get_alignment_row_dists(np.fliplr(P1), np.fliplr(P2))
    dists22 = get_alignment_row_dists(np.fliplr(P2), np.fliplr(P1))
    return np.concatenate((dists11, dists12, dists21, dists22))

def get_hist(dists):
    """
    Return a histogram of distances
    """
    hist = {}
    for d in dists:
        if d in hist:
            hist[d] += 1
        else:
            hist[d] = 1
    return hist

def get_inverse_fn_equally_sampled(t, x):
    import scipy.interpolate as interp
    N = len(t)
    t2 = np.linspace(np.min(x), np.max(x), N)
    try:
        res = interp.spline(x, t, t2)
        return res
    except:
        return t

def get_warp_dictionary(N, plotPaths = False):
    t = np.linspace(0, 1, N)
    D = []
    #Polynomial
    if plotPaths:
        import matplotlib.pyplot as plt
        plt.subplot(131)
        plt.title('Polynomial')
    for p in range(-4, 6):
        tp = p
        if tp < 0:
            tp = -1.0/tp
        x = t**(tp**1)
        D.append(x)
        if plotPaths:
            import matplotlib.pyplot as plt
            plt.plot(x)
    #Exponential / Logarithmic
    if plotPaths:
        import matplotlib.pyplot as plt
        plt.subplot(132)
        plt.title('Exponential / Logarithmic')
    for p in range(2, 6):
        t = np.linspace(1, p**p, N)
        x = np.log(t)
        x = x - np.min(x)
        x = x/np.max(x)
        t = t/np.max(t)
        x2 = get_inverse_fn_equally_sampled(t, x)
        x2 = x2 - np.min(x2)
        x2 = x2/np.max(x2)
        #D.append(x)
        #D.append(x2)
        if plotPaths:
            import matplotlib.pyplot as plt
            plt.plot(x)
            plt.plot(x2)
    #Hyperbolic Tangent
    if plotPaths:
        import matplotlib.pyplot as plt
        plt.subplot(133)
        plt.title('Hyperbolic Tangent')
    for p in range(2, 5):
        t = np.linspace(-2, p, N)
        x = np.tanh(t)
        x = x - np.min(x)
        x = x/np.max(x)
        t = t/np.max(t)
        x2 = get_inverse_fn_equally_sampled(t, x)
        x2 = x2 - np.min(x2)
        x2 = x2/np.max(x2)
        D.append(x)
        D.append(x2)
        if plotPaths:
            import matplotlib.pyplot as plt
            plt.plot(x)
            plt.plot(x2)
    D = np.array(D)
    return D

def get_warping_path_dict(D, k, doPlot = False):
    """
    Return a warping path made up of k elements
    drawn from dictionary D
    """
    N = D.shape[0]
    dim = D.shape[1]
    idxs = np.random.permutation(N)[0:k]
    weights = np.zeros(N)
    weights[idxs] = np.random.rand(k)
    weights = weights / np.sum(weights)
    res = weights.dot(D)
    res = res - np.min(res)
    res = res/np.max(res)
    if doPlot:
        import matplotlib.pyplot as plt
        plt.plot(res)
        for idx in idxs:
            plt.plot(np.arange(dim), D[idx, :], linestyle='--')
        plt.title('Constructed Warping Path')
    return res

def get_interpolated_euclidean_timeseries(X, t):
    import scipy.interpolate as interp
    M = X.shape[0]
    d = X.shape[1]
    t0 = np.linspace(0, 1, M)
    dix = np.arange(d)
    f = interp.interp2d(dix, t0, X, kind='linear')
    Y = f(dix, t)
    return Y


import numpy as np
import matplotlib.pyplot as plt
import warnings

def load_audio(filename, sr = 44100):
    """
    Load an audio waveform from a file.  Try to use ffmpeg
    to convert it to a .wav file so scipy's fast wavfile loader
    can work.  Otherwise, fall back to the slower librosa
    Parameters
    ----------
    filename: string
        Path to audio file to load
    sr: int
        Sample rate to use
    Returns
    -------
    y: ndarray(N)
        Audio samples
    sr: int
        The sample rate that was actually used
    """
    try:
        # First, try a faster version of loading audio
        from scipy.io import wavfile
        import subprocess
        import os
        FFMPEG_BINARY = "ffmpeg"
        wavfilename = "%s.wav"%filename
        if os.path.exists(wavfilename):
            os.remove(wavfilename)
        subprocess.call([FFMPEG_BINARY, "-i", filename, "-ar", "%i"%sr, "-ac", "1", wavfilename], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        _, y = wavfile.read(wavfilename)
        y = y/2.0**15
        os.remove(wavfilename)
        return y, sr
    except:
        # Otherwise, fall back to librosa
        warnings.warn("Falling back to librosa for audio reading, which may be slow for long audio files")
        import librosa
        return librosa.load(filename, sr=sr)

def get_DLNC0(x, sr, hop_length, lag=10, do_plot=False):
    """
    Compute decaying locally adaptive normalize C0 (DLNC0) features
    Parameters
    ----------
    x: ndarray(N)
        Audio samples
    sr: int
        Sample rate
    hop_length: int
        Hop size between windows
    lag: int
        Number of lags to include
    Returns
    -------
    X: ndarray(n_win, 12)
        The DLNC0 features
    """
    from scipy.ndimage.filters import gaussian_filter1d as gf1d
    from scipy.ndimage.filters import maximum_filter1d
    import librosa
    X = np.abs(librosa.cqt(x, sr=sr, hop_length=hop_length, bins_per_octave=12))
    # Half-wave rectify discrete derivative
    #X = librosa.amplitude_to_db(X, ref=np.max)
    #X[:, 0:-1] = X[:, 1::] - X[:, 0:-1]
    X = gf1d(X, 5, axis=1, order = 1)
    X[X < 0] = 0
    # Retain peaks
    XLeft = X[:, 0:-2]
    XRight = X[:, 2::]
    mask = np.zeros_like(X)
    mask[:, 1:-1] = (X[:, 1:-1] > XLeft)*(X[:, 1:-1] > XRight)
    X[mask < 1] = 0
    # Fold into octave
    n_octaves = int(X.shape[0]/12)
    X2 = np.zeros((12, X.shape[1]), dtype=X.dtype)
    for i in range(n_octaves):
        X2 += X[i*12:(i+1)*12, :]
    X = X2
    # Compute norms
    if do_plot:
        import librosa.display
        plt.subplot(411)
        librosa.display.specshow(X, sr=sr, x_axis='time', y_axis='chroma')
    norms = np.sqrt(np.sum(X**2, 0))
    if do_plot:
        plt.subplot(412)
        plt.plot(norms)
    norms = maximum_filter1d(norms, size=int(2*sr/hop_length))
    if do_plot:
        import librosa.display
        plt.plot(norms)
        plt.subplot(413)
        X = X/norms[None, :]
        librosa.display.specshow(X, sr=sr, x_axis='time', y_axis='chroma')
    # Apply LNCO
    decays = np.linspace(0, 1, lag+1)[1::]
    decays = np.sqrt(decays[::-1])
    XRet = np.zeros_like(X)
    M = X.shape[1]-lag+1
    for i in range(lag):
        XRet[:, i:i+M] += X[:, 0:M]*decays[i]
    if do_plot:
        plt.subplot(414)
        librosa.display.specshow(XRet, sr=sr, x_axis='time', y_axis='chroma')
        plt.show()
    return XRet

def get_mixed_DLNC0_CENS(x, sr, hop_length, lam=0.1):
    """
    Concatenate DLNC0 to CENS
    Parameters
    ----------
    x: ndarray(N)
        Audio samples
    sr: int
        Sample rate
    hop_length: int
        Hop size between windows
    lam: float
        The coefficient in front of the CENS features
    Returns
    -------
    X: ndarray(n_win, 24)
        DLNC0 features along the first 12 columns, 
        weighted CENS along the next 12 columns
    """
    import librosa
    X1 = get_DLNC0(x, sr, hop_length).T
    X2 = lam*librosa.feature.chroma_cens(y=x, sr=sr, hop_length=hop_length).T
    return np.concatenate((X1, X2), axis=1)

def get_mfcc_mod(x, sr, hop_length, n_mfcc=120, drop=20, n_fft = 2048):
    """
    Compute the mfcc_mod features, as described in Gadermaier 2019
    Parameters
    ----------
    x: ndarray(N)
        Audio samples
    sr: int
        Sample rate
    hop_length: int
        Hop size between windows
    n_mfcc: int
        Number of mfcc coefficients to compute
    drop: int
        Index under which to ignore coefficients
    n_fft: int
        Number of fft points to use in each window
    Returns
    -------
    X: ndarray(n_win, n_mfcc-drop)
        The mfcc-mod features
    """
    import skimage.transform
    import librosa
    X = librosa.feature.mfcc(y=x, sr=sr, hop_length=hop_length, n_mfcc = n_mfcc, n_fft=n_fft, htk=True)
    X = X[drop::, :].T
    return X

def stretch_audio(x1, x2, sr, path, hop_length):
    """
    Wrap around pyrubberband to warp one audio stream
    to another, according to some warping path
    Parameters
    ----------
    x1: ndarray(M)
        First audio stream
    x2: ndarray(N)
        Second audio stream
    sr: int
        Sample rate
    path: ndarray(P, 2)
        Warping path, in units of windows
    hop_length: int
        The hop length between windows
    Returns
    -------
    x3: ndarray(N, 2)
        The synchronized audio.  x2 is in the right channel,
        and x1 stretched to x2 is in the left channel
    """
    from .alignmenttools import make_path_strictly_increase
    import pyrubberband as pyrb
    print("Stretching...")
    path = make_path_strictly_increase(path)
    path *= hop_length
    path = [(row[0], row[1]) for row in path]
    path.append((x1.size, x2.size))
    x3 = np.zeros((x2.size, 2))
    x3[:, 1] = x2
    x1_stretch = pyrb.timemap_stretch(x1, sr, path)
    x3[:, 0] = x1_stretch
    return x3


import numpy as np
from .alignmenttools import get_diag_len, get_diag_indices, update_alignment_metadata
import warnings

def dtw_brute(X, Y, debug=False):
    """
    Compute brute force dynamic time warping between two 
    time-ordered point clouds in Euclidean space, using 
    cython on the backend
    Parameters
    ----------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    debug: boolean
        Whether to keep track of debugging information
    Returns
    -------
    {
        'cost': float
            The optimal cost of the alignment (if computation didn't stop prematurely),
        'U'/'L'/'UL': ndarray(M, N)
            The choice matrices (if debugging),
        'S': ndarray(M, N)
            The accumulated cost matrix (if debugging)
    }
    """
    from dynseqalign import DTW
    if not X.dtype == np.float32:
        warnings.warn("X is not 32-bit, so creating 32-bit version")
        X = np.array(X, dtype=np.float32)
    if not Y.dtype == np.float32:
        warnings.warn("Y is not 32-bit, so creating 32-bit version")
        Y = np.array(Y, dtype=np.float32)
    if X.shape[0] < X.shape[1]:
        warnings.warn("X {} has more columns than rows; did you mean to transpose?".format(X.shape))
    if Y.shape[0] < Y.shape[1]:
        warnings.warn("Y {} has more columns than rows; did you mean to transpose?".format(Y.shape))
    return DTW(X, Y, int(debug))

def dtw_brute_backtrace(X, Y, debug=False):
    """
    Compute dynamic time warping between two time-ordered
    point clouds in Euclidean space, using cython on the 
    backend.  Then, trace back through the matrix of backpointers
    to extract an alignment path
    Parameters
    ----------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    debug: boolean
        Whether to keep track of debugging information
    Returns
    -------
    path (If not debugging): ndarray(K, 2)
        The warping path
    
    If debugging
    {
        'cost': float
            The optimal cost of the alignment (if computation didn't stop prematurely),
        'U'/'L'/'UL': ndarray(M, N)
            The choice matrices (if debugging),
        'S': ndarray(M, N)
            The accumulated cost matrix (if debugging)
        'path': ndarray(K, 2)
            The warping path
    }
    """
    res = dtw_brute(X, Y, debug)
    res['P'] = np.asarray(res['P'])
    if debug:
        for key in ['U', 'L', 'UL', 'S']:
            res[key] = np.asarray(res[key])
    i = X.shape[0]-1
    j = Y.shape[0]-1
    path = [[i, j]]
    step = [[0, -1], [-1, 0], [-1, -1]] # LEFT, UP, DIAG
    while not(path[-1][0] == 0 and path[-1][1] == 0):
        s = step[res['P'][i, j]]
        i += s[0]
        j += s[1]
        path.append([i, j])
    path.reverse()
    if debug:
        res['path'] = path
        return res
    return path
    

def dtw_diag(X, Y, k_save = -1, k_stop = -1, box = None, reverse=False, debug=False, metadata=None):
    """
    A CPU version of linear memory diagonal DTW
    Parameters
    ----------
    X: ndarray(M, d)
        An M-dimensional Euclidean point cloud
    Y: ndarray(N, d)
        An N-dimensional Euclidean point cloud
    k_save: int
        Index of the diagonal d2 at which to save d0, d1, and d2
    k_stop: int
        Index of the diagonal d2 at which to stop computation
    box: list
        A list of [startx, endx, starty, endy]
    reverse: boolean
        Whether we're going in reverse
    debug: boolean
        Whether to save the accumulated cost matrix
    metadata: dictionary
        A dictionary for storing information about the computation
    Returns
    -------
    {
        'cost': float
            The optimal cost of the alignment (if computation didn't stop prematurely),
        'U'/'L'/'UL': ndarray(M, N)
            The choice matrices (if debugging),
        'd0'/'d1'/'d2':ndarray(min(M, N))
            The saved rows if a save index was chosen,
        'csm0'/'csm1'/'csm2':ndarray(min(M, N))
            The saved cross-similarity distances if a save index was chosen
    }
    """
    from dynseqalign import DTW_Diag_Step
    if not box:
        box = [0, X.shape[0]-1, 0, Y.shape[0]-1]
    M = box[1] - box[0] + 1
    N = box[3] - box[2] + 1
    box = np.array(box, dtype=np.int32)

    # Debugging info
    U = np.zeros((1, 1), dtype=np.float32)
    L = np.zeros_like(U)
    UL = np.zeros_like(U)
    S = np.zeros_like(U)
    if debug:
        U = np.zeros((M, N), dtype=np.float32)
        L = np.zeros_like(U)
        UL = np.zeros_like(U)
        S = np.zeros_like(U)
    
    # Diagonals
    diagLen = min(M, N)
    d0 = np.zeros(diagLen, dtype=np.float32)
    d1 = np.zeros_like(d0)
    d2 = np.zeros_like(d0)
    # Distances between points along diagonals
    csm0 = np.zeros_like(d0)
    csm1 = np.zeros_like(d1)
    csm2 = np.zeros_like(d2)
    csm0len = diagLen
    csm1len = diagLen
    csm2len = diagLen

    # Loop through diagonals
    res = {}
    for k in range(M+N-1):
        DTW_Diag_Step(d0, d1, d2, csm0, csm1, csm2, X, Y, diagLen, box, int(reverse), k, int(debug), U, L, UL, S)
        csm2len = get_diag_len(box, k)
        if metadata:
            update_alignment_metadata(metadata, csm2len)
        if k == k_save:
            res['d0'] = d0.copy()
            res['csm0'] = csm0.copy()
            res['d1'] = d1.copy()
            res['csm1'] = csm1.copy()
            res['d2'] = d2.copy()
            res['csm2'] = csm2.copy()
        if k == k_stop:
            break
        # Shift diagonals
        temp = d0
        d0 = d1
        d1 = d2
        d2 = temp
        temp = csm0
        csm0 = csm1
        csm1 = csm2
        csm2 = temp
        temp = csm0len
        csm0len = csm1len
        csm1len = csm2len
        csm2len = temp
    res['cost'] = d2[0] + csm2[0]
    if debug:
        res['U'] = U
        res['L'] = L
        res['UL'] = UL
        res['S'] = S
    return res

def linmdtw(X, Y, box = None, min_dim = 500, do_gpu = True, metadata = None):
    """
    Linear memory exact, parallelizable DTW
    Parameters
    ----------
    X: ndarray(N1, d)
        An N1-dimensional Euclidean point cloud
    Y: ndarray(N2, d)
        An N2-dimensional Euclidean point cloud
    min_dim: int
        If one of the dimensions of the rectangular region
        to the left or to the right is less than this number,
        then switch to brute force CPU
    do_gpu: boolean
        If true, use the GPU diagonal DTW function as a subroutine.
        Otherwise, use the CPU version.  Both are linear memory, but 
        the GPU will go faster for larger synchronization problems
    metadata: dictionary
        A dictionary for storing information about the computation
    Returns
    -------
    path: ndarray(K, 2)
        The optimal warping path
    """
    if not X.dtype == np.float32:
        warnings.warn("X is not 32-bit, so creating 32-bit version")
        X = np.array(X, dtype=np.float32)
    if not Y.dtype == np.float32:
        warnings.warn("Y is not 32-bit, so creating 32-bit version")
        Y = np.array(Y, dtype=np.float32)
    if X.shape[0] < X.shape[1]:
        warnings.warn("X {} has more columns than rows; did you mean to transpose?".format(X.shape))
    if Y.shape[0] < Y.shape[1]:
        warnings.warn("Y {} has more columns than rows; did you mean to transpose?".format(Y.shape))
    dtw_diag_fn = dtw_diag
    if do_gpu:
        from .dtwgpu import DTW_GPU_Initialized, init_gpu, dtw_diag_gpu
        if not DTW_GPU_Initialized:
            init_gpu()
        from .dtwgpu import DTW_GPU_Failed
        if DTW_GPU_Failed:
            warnings.warn("Falling back to CPU")
            do_gpu = False
        else:
            dtw_diag_fn = dtw_diag_gpu
    if not box:
        box = [0, X.shape[0]-1, 0, Y.shape[0]-1]
    M = box[1]-box[0]+1
    N = box[3]-box[2]+1

    # Stopping condition, revert to CPU
    if M < min_dim or N < min_dim:
        if metadata:
            metadata['totalCells'] += M*N
        path = dtw_brute_backtrace(X[box[0]:box[1]+1, :], Y[box[2]:box[3]+1, :])
        for p in path:
            p[0] += box[0]
            p[1] += box[2]
        return path
    
    # Otherwise, proceed with recursion
    K = M + N - 1
    # Do the forward computation
    k_save = int(np.ceil(K/2.0))
    res1 = dtw_diag_fn(X, Y, k_save=k_save, k_stop=k_save, box=box, metadata=metadata)

    # Do the backward computation
    k_save_rev = k_save
    if K%2 == 0:
        k_save_rev += 1
    res2 = dtw_diag_fn(X, Y, k_save=k_save_rev, k_stop=k_save_rev, box=box, reverse=True, metadata=metadata)
    res2['d0'], res2['d2'] = res2['d2'], res2['d0']
    res2['csm0'], res2['csm2'] = res2['csm2'], res2['csm0']
    # Chop off extra diagonal elements
    for res in res1, res2:
        for i in range(3):
            res['d%i'%i] = res['d%i'%i][0:res['csm%i'%i].size]
    # Line up the reverse diagonals
    for d in ['d0', 'd1', 'd2', 'csm0', 'csm1', 'csm2']:
        res2[d] = res2[d][::-1]
    
    # Find the min cost over the three diagonals and split on that element
    min_cost = np.inf
    min_idxs = []
    for k in range(3):
        dleft = res1['d%i'%k]
        dright = res2['d%i'%k]
        csmright = res2['csm%i'%k]
        diagsum = dleft + dright + csmright
        idx = np.argmin(diagsum)
        if diagsum[idx] < min_cost:
            min_cost = diagsum[idx]
            i, j = get_diag_indices(X.shape[0], Y.shape[0], k_save-2+k, box)
            min_idxs = [i[idx], j[idx]]

    # Recursively compute left paths
    left_path = []
    box_left = [box[0], min_idxs[0], box[2], min_idxs[1]]
    left_path = linmdtw(X, Y, box_left, min_dim, do_gpu, metadata)

    # Recursively compute right paths
    right_path = []
    box_right = [min_idxs[0], box[1], min_idxs[1], box[3]]
    right_path = linmdtw(X, Y, box_right, min_dim, do_gpu, metadata)
    
    return left_path + right_path[1::]

    


import sys
import os
import platform

from setuptools import setup
from setuptools.extension import Extension

# Ensure Cython is installed before we even attempt to install linmdtw
try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext
except:
    print("You don't seem to have Cython installed. Please get a")
    print("copy from www.cython.org or install it with `pip install Cython`")
    sys.exit(1)

## Get version information from _version.py
import re
VERSIONFILE="linmdtw/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# Use README.md as the package long description  
with open('README.md') as f:
    long_description = f.read()

class CustomBuildExtCommand(build_ext):
    """ This extension command lets us not require numpy be installed before running pip installing
        build_ext command for use when numpy headers are needed.
    """

    def run(self):
        # Import numpy here, only when headers are needed
        import numpy
        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())
        # Call original build_ext command
        build_ext.run(self)

extra_compile_args = []
extra_link_args = []

if platform.system() == "Darwin":
    extra_compile_args.extend([
        "-mmacosx-version-min=10.9"
    ])
    extra_link_args.extend([
        "-mmacosx-version-min=10.9"
    ])

ext_modules = Extension(
    "dynseqalign",
    sources=["linmdtw/dynseqalign.pyx"],
    define_macros=[
    ],
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c++"
)


setup(
    name="linmdtw",
    version=verstr,
    description="A linear memory, exact, parallelizable algorithm for dynamic time warping in arbitrary Euclidean spaces",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Anonymous",
    author_email="ctralie@alumni.princeton.edu",
    license='Apache2',
    packages=['linmdtw'],
    ext_modules=cythonize(ext_modules, include_path=['linmdtw']),
    install_requires=[
        'Cython',
        'numpy'
    ],
    extras_require={
        'testing': [ # `pip install -e ".[testing]"``
            'pytest'  
        ],
        'docs': [ # `pip install -e ".[docs]"`
            'sktda_docs_config'
        ],
        'examples': []
    },
    cmdclass={'build_ext': CustomBuildExtCommand},
    python_requires='>=3.6',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    keywords='dynamic time warping, fast dtw, synchronization'
)



import pytest
import numpy as np
import linmdtw

class TestLibrary:
    # Does the library install in scope? Are the objects in scope?
    def test_import(self):
        import linmdtw
        from linmdtw import linmdtw, dtw_brute, dtw_brute_backtrace, fastdtw, mrmsdtw
        assert 1


class TestDTW:
    def test_input_type_warning(self):
        np.random.seed(0)
        X = np.random.rand(10, 3)
        with pytest.warns(UserWarning, match="is not 32-bit, so creating 32-bit version") as w:
            linmdtw.linmdtw(X, X)
        with pytest.warns(UserWarning, match="is not 32-bit, so creating 32-bit version") as w:
            linmdtw.dtw_brute_backtrace(X, X)

    def test_dimension_warning(self):
        np.random.seed(0)
        X = np.random.rand(3, 10)
        with pytest.warns(UserWarning, match="has more columns than rows") as w:
            linmdtw.linmdtw(X, X)
        with pytest.warns(UserWarning, match="has more columns than rows") as w:
            linmdtw.dtw_brute_backtrace(X, X)

    def test_cpu_vs_gpu(self):
        N = 300
        t1 = np.linspace(0, 1, N)
        t2 = t1**2
        X = np.zeros((N, 2), dtype=np.float32)
        X[:, 0] = np.cos(2*np.pi*t1)
        X[:, 1] = np.sin(4*np.pi*t1)
        Y = np.zeros_like(X)
        Y[:, 0] = np.cos(2*np.pi*t2)
        Y[:, 1] = np.sin(4*np.pi*t2)
        path1 = linmdtw.linmdtw(X, Y)
        path1 = np.array(path1)
        path2 = linmdtw.dtw_brute_backtrace(X, Y)
        path2 = np.array(path2)
        assert(np.allclose(path1, path2))



import numpy as np
import matplotlib.pyplot as plt
import scipy.io as sio
from scipy import sparse
import time
import dynseqalign
from .dtw import dtw_brute_backtrace, linmdtw

def fill_block(A, p, radius, val):
    """
    Fill a square block with values
    Parameters
    ----------
    A: ndarray(M, N) or sparse(M, N)
        The array to fill
    p: list of [i, j]
        The coordinates of the center of the box
    radius: int
        Half the width of the box
    val: float
        Value to fill in
    """
    i1 = max(p[0]-radius, 0)
    i2 = min(p[0]+radius, A.shape[0]-1)
    j1 = max(p[1]-radius, 0)
    j2 = min(p[1]+radius, A.shape[1]-1)
    A[i1:i2+1, j1:j2+1] = val

def fastdtw(X, Y, radius, debug=False, level = 0, do_plot=False):
    """
    An implementation of [1]
    [1] FastDTW: Toward Accurate Dynamic Time Warping in Linear Time and Space
                Stan Salvador and Philip Chan
    Parameters
    ----------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    radius: int
        Radius of the l-infinity box that determines sparsity structure
        at each level
    debug: boolean
        Whether to keep track of debugging information
    level: int
        An int for keeping track of the level of recursion
    do_plot: boolean
        Whether to plot the warping path at each level and save to image files
    Returns
    -------
    path: ndarray(K, 2)
        The warping path
    """
    minTSsize = radius + 2
    M = X.shape[0]
    N = Y.shape[0]
    X = np.ascontiguousarray(X)
    Y = np.ascontiguousarray(Y)
    if M < radius or N < radius:
        return dtw_brute_backtrace(X, Y)
    # Recursive step
    path = fastdtw(X[0::2, :], Y[0::2, :], radius, debug, level+1, do_plot)
    if type(path) is dict:
        path = path['path']
    path = np.array(path)
    path *= 2
    tic = time.time()
    S = sparse.lil_matrix((M, N))
    P = sparse.lil_matrix((M, N), dtype=int)
    Occ = sparse.lil_matrix((M, N))

    ## Step 1: Figure out the indices of the occupied cells
    for p in path:
        fill_block(Occ, p, radius, 1)
    I, J = Occ.nonzero()
    # Sort cells in raster order
    idx = np.argsort(J)
    I = I[idx]
    J = J[idx]
    idx = np.argsort(I, kind='stable')
    I = I[idx]
    J = J[idx]

    ## Step 2: Find indices of left, up, and diag neighbors.  
    # All neighbors must be within bounds *and* within sparse structure
    # Make idx M+1 x N+1 so -1 will wrap around to 0
    # Make 1-indexed so all valid entries have indices > 0
    idx = sparse.coo_matrix((np.arange(I.size)+1, (I, J)), shape=(M+1, N+1)).tocsr()
    # Left neighbors
    left = np.array(idx[I, J-1], dtype=np.int32).flatten()
    left[left <= 0] = -1
    left -= 1
    # Up neighbors
    up = np.array(idx[I-1, J], dtype=np.int32).flatten()
    up[up <= 0] = -1
    up -= 1
    # Diag neighbors
    diag = np.array(idx[I-1, J-1], dtype=np.int32).flatten()
    diag[diag <= 0] = -1
    diag -= 1

    ## Step 3: Pass information to cython for dynamic programming steps
    S = np.zeros(I.size, dtype=np.float32) # Dyn prog matrix
    P = np.zeros(I.size, dtype=np.int32) # Path pointer matrix
    dynseqalign.FastDTW_DynProg_Step(X, Y, I, J, left, up, diag, S, P)
    P = sparse.coo_matrix((P, (I, J)), shape=(M, N)).tocsr()
    if debug or do_plot:
        S = sparse.coo_matrix((S, (I, J)), shape=(M, N)).tocsr()
    
    # Step 4: Do backtracing
    i = M-1
    j = N-1
    path = [[i, j]]
    step = [[0, -1], [-1, 0], [-1, -1]] # LEFT, UP, DIAG
    while not(path[-1][0] == 0 and path[-1][1] == 0):
        s = step[P[i, j]]
        i += s[0]
        j += s[1]
        path.append([i, j])
    path.reverse()
    
    if do_plot:
        plt.figure(figsize=(8, 8))
        plt.imshow(S.toarray())
        path = np.array(path)
        plt.scatter(path[:, 1], path[:, 0], c='C1')
        plt.title("Level {}".format(level))
        plt.savefig("%i.png"%level, bbox_inches='tight')

    if debug:
        return {'path':path, 'S':S, 'P':P}
    else:
        return path

def get_box_area(a1, a2):
    """
    Get the area of a box specified by two anchors
    Parameters
    ----------
    a1: list(2)
        Row/column of first anchor
    a2: list(2)
        Row/column of second anchor
    Returns
    -------
    Area of box determined by these two anchors
    """
    m = a2[0]-a1[0]+1
    n = a2[1]-a1[1]+1
    return m*n

def mrmsdtw(X, Y, tau, debug=False, refine=True):
    """
    An implementation of the approximate, memory-restricted
    multiscale DTW technique from [2]
    [2] "Memory-Restricted Multiscale Dynamic Time Warping"
    Thomas Praetzlich, Jonathan Driedger and Meinard Mueller
    Parameters
    ----------
    X: ndarray(M, d)
        A d-dimensional Euclidean point cloud with M points
    Y: ndarray(N, d)
        A d-dimensional Euclidean point cloud with N points
    tau: int
        The max amount of cells to be in memory at any given
        time
    debug: boolean
        Whether to keep track of debugging information
    refine: boolean
        Whether to do refinement with the "white anchors"
    Returns
    -------
    path: ndarray(K, 2)
        The warping path
    """
    X = np.ascontiguousarray(X)
    Y = np.ascontiguousarray(Y)
    M = X.shape[0]
    N = Y.shape[0]
    if M*N < tau:
        # If the matrix is already within the memory bounds, simply
        # return DTW
        return dtw_brute_backtrace(X, Y)

    ## Step 1: Perform DTW at the coarse level
    # Figure out the subsampling factor for the
    # coarse alignment based on memory requirements
    d = int(np.ceil(np.sqrt(M*N/tau)))
    X2 = np.ascontiguousarray(X[0::d, :])
    Y2 = np.ascontiguousarray(Y[0::d, :])
    anchors = dtw_brute_backtrace(X2, Y2)
    anchors = (np.array(anchors)*d).tolist()
    if anchors[-1][0] < M-1 or anchors[-1][1] < N-1:
        anchors.append([M-1, N-1])
    
    ## Step 2: Subdivide anchors if necessary to keep
    # within memory bounds
    idx = 0
    while idx < len(anchors)-1:
        a1 = anchors[idx]
        a2 = anchors[idx+1]
        if get_box_area(a1, a2) > tau:
            # Subdivide cell
            i = int((a1[0]+a2[0])/2)
            j = int((a1[1]+a2[1])/2)
            anchors = anchors[0:idx+1] + [[i, j]] + anchors[idx+1::]
        else:
            # Move on
            idx += 1
    
    ## Step 3: Do alignments in each block
    path = []
    # Keep track of the "black anchor" indices in the path
    banchors_idx = [0]
    for i in range(len(anchors)-1):
        a1 = anchors[i]
        a2 = anchors[i+1]
        box = [a1[0], a2[0], a1[1], a2[1]]
        pathi = linmdtw(X, Y, box=box)
        path += pathi[0:-1]
        banchors_idx.append(len(path)-1)
    path += [[M-1, N-1]]
    if not refine:
        return path
    
    ## Step 4: Come up with the set of "white anchors"
    # First choose them to be at the center of each block
    wanchors_idx = []
    for idx in range(len(banchors_idx)-1):
        wanchors_idx.append([int(0.5*(banchors_idx[idx]+banchors_idx[idx+1]))]*2)
    # Split anchor positions if the blocks are too big
    for i in range(len(wanchors_idx)-1):
        a1 = path[wanchors_idx[i][-1]]
        a2 = path[wanchors_idx[i+1][0]]
        while get_box_area(a1, a2) > tau:
            # Move the anchors towards each other
            wanchors_idx[i][-1] += 1
            wanchors_idx[i+1][0] -= 1
            a1 = path[wanchors_idx[i][-1]]
            a2 = path[wanchors_idx[i+1][0]]
    ## Step 5: Do DTW between white anchors and splice path together
    pathret = path[0:wanchors_idx[0][0]+1]
    for i in range(len(wanchors_idx)-1):
        a1 = path[wanchors_idx[i][-1]]
        a2 = path[wanchors_idx[i+1][0]]
        box = [a1[0], a2[0], a1[1], a2[1]]
        pathi = linmdtw(X, Y, box=box)
        pathret += pathi[0:-1]
        # If there's a gap in between this box and 
        # the next one, use the path from before
        i1 = wanchors_idx[i+1][0]
        i2 = wanchors_idx[i+1][1]
        if i1 != i2:
            pathret += path[i1:i2]
    i1 = wanchors_idx[-1][-1]
    pathret += path[i1::]
    return pathret